/*
  AudioGeneratorRAW
  Audio output generator that reads 8 and 16-bit RAW files
  
  Copyright (C) 2017  Earle F. Philhower, III

  This program is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "AudioGeneratorRAW.h"

AudioGeneratorRAW::AudioGeneratorRAW()
{
  running = false;
  file = NULL;
  output = NULL;
  buffSize = 2048; // 128
  buff = NULL;
  buffPtr = 0;
  buffLen = 0;
  beginPos = 0;
  loopaudio = false;
}

AudioGeneratorRAW::~AudioGeneratorRAW()
{
  free(buff);
  buff = NULL;
}

bool AudioGeneratorRAW::stop()
{
  if (!running) return true;
  running = false;
  free(buff);
  buff = NULL;
  output->stop();
  return file->close();
}

bool AudioGeneratorRAW::endofdata()
{
  if (!loopaudio) return stop();
  resetBegin();
  return true;
}

bool AudioGeneratorRAW::isRunning()
{
  return running;
}

// Handle buffered reading, reload each time we run out of data
bool AudioGeneratorRAW::GetBufferedData(int bytes, void *dest)
{
  if (!running) return false; // Nothing to do here!
  uint8_t *p = reinterpret_cast<uint8_t*>(dest);
  while (bytes--) {
    // Potentially load next batch of data...
    if (buffPtr >= buffLen) {
      buffPtr = 0;
      uint32_t toRead = availBytes > buffSize ? buffSize : availBytes;
      buffLen = file->read( buff, toRead );
      availBytes -= buffLen;
    }
    if (buffPtr >= buffLen)
      return false; // No data left!
    *(p++) = buff[buffPtr++];
  }
  return true;
}

bool AudioGeneratorRAW::loop()
{
  if (!running) goto done; // Nothing to do here!
  
  // First, try and push in the stored sample.  If we can't, then punt and try later
  if (!output->ConsumeSample(lastSample)) goto done; // Can't send, but no error detected

  // Try and stuff the buffer one sample at a time
  do
  {
    if (bitsPerSample == 8) {
      uint8_t l, r;
      if (!GetBufferedData(1, &l)) endofdata();
      if (channels == 2) {
        if (!GetBufferedData(1, &r)) endofdata();
      } else {
        r = 0;
      }
      lastSample[AudioOutput::LEFTCHANNEL] = l;
      lastSample[AudioOutput::RIGHTCHANNEL] = r;
    } else if (bitsPerSample == 16) {
      if (!GetBufferedData(2, &lastSample[AudioOutput::LEFTCHANNEL])) endofdata();
      if (channels == 2) {
        if (!GetBufferedData(2, &lastSample[AudioOutput::RIGHTCHANNEL])) endofdata();
      } else {
        lastSample[AudioOutput::RIGHTCHANNEL] = 0;
      }
    }
  } while (running && output->ConsumeSample(lastSample));

done:
  file->loop();
  output->loop();

  return running;
}

bool AudioGeneratorRAW::resetBegin()
{
  Serial.print("AudioGeneratorRAW::resetBegin\n");
  file->seek(beginPos, SEEK_SET);
  buffPtr = 0;
  buffLen = 0;
  availBytes = FileavailBytes;
  return true;
}

void AudioGeneratorRAW::setLoopAudio()
{
loopaudio = true;
}

bool AudioGeneratorRAW::begin(AudioFileSource *source, AudioOutput *output)
{
  if (!source) {
    Serial.printf_P(PSTR("AudioGeneratorRAW::begin: failed: invalid source\n"));
    return false;
  }
  file = source;
  if (!output) {
    Serial.printf_P(PSTR("AudioGeneratorRAW::begin: invalid output\n"));
    return false;
  }
  this->output = output;
  if (!file->isOpen()) {
    Serial.printf_P(PSTR("AudioGeneratorRAW::begin: file not open\n"));
    return false;
  } // Error

  availBytes = file->getSize();
  FileavailBytes = availBytes;
  buffPtr = 0;
  buffLen = 0;
  beginPos = 0;
  bitsPerSample = 16;
  channels = 1;

  Serial.print("file size: ");
  Serial.println(availBytes);
  
  // Now set up the buffer or fail
  buff = reinterpret_cast<uint8_t *>(malloc(buffSize));
  if (!buff) {
    Serial.printf_P(PSTR("AudioGeneratorRAW::ReadWAVInfo: cannot read WAV, failed to set up buffer \n"));
    return false;
  };

  if (!output->begin()) {
    Serial.printf_P(PSTR("AudioGeneratorRAW::begin: output's begin did not return true\n"));
    return false;
  }

  running = true;

  return true;
}
